package autoeventid;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.Writer;
import java.util.ArrayList;

public class Main
{
  static ArrayList<String> parsedNPCNames;
  static ArrayList<byte[]> parsedNPCIDs;
  static String fileout;
  
  public static void main(String[] args)
  {
    System.out.println("Npc Event Id Parser for FINAL FANTASY XI");
    String root = "";
    fileout = "";
    
    parsedNPCNames = new ArrayList();
    parsedNPCIDs = new ArrayList();
    ArrayList<Event> zonelist = new ArrayList();
    try
    {
      System.out.println("Reading fileinput.txt...");
      FileInputStream fis = new FileInputStream("fileinput.txt");
      DataInputStream in = new DataInputStream(fis);
      BufferedReader br = new BufferedReader(new InputStreamReader(in));
      String str = "";
      while (br.ready())
      {
        str = br.readLine();
        String mobentry = "";
        String evententry = "";
        String nameentry = "";
        
        int i_comment = str.indexOf("//", 0);
        if (i_comment != -1) {
          str = str.substring(0, i_comment);
        }
        if (str.length() != 0)
        {
          int i_mob = str.indexOf('#');
          int i_evt = str.indexOf('@');
          int i_name = str.indexOf('%');
          if (i_mob != -1) {
            mobentry = str.substring(i_mob + 1, str.indexOf(' ', i_mob));
          }
          if (i_evt != -1) {
            evententry = str.substring(i_evt + 1, str.indexOf(' ', i_evt));
          }
          if (i_name != -1) {
            nameentry = str.substring(i_name + 1);
          }
          if (str.substring(0, 5).equals("root=")) {
            root = str.substring(5);
          }
        }
        if (nameentry.length() > 0)
        {
          Event e = new Event(nameentry);
          e.setEventPath(evententry);
          e.setMobPath(mobentry);
          if (e.hasMissingPaths()) {
            e.autoCorrectPaths();
          }
          zonelist.add(e);
        }
      }
    }
    catch (Exception e)
    {
      System.out.println("ERROR: File Input: " + e);
    }
    if (root.length() == 0)
    {
      System.out.println("No root path directory found. Please enter the FFXI directory:");
      BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
      try
      {
        root = in.readLine();
      }
      catch (Exception e)
      {
        System.out.println("ERROR: console input: " + e);
      }
    }
    System.out.println("Starting Monster ID/Event ID matching...");
    for (int i = 0; i < zonelist.size(); i++)
    {
      if (!parsedNPCNames.isEmpty()) {
        parsedNPCNames.clear();
      }
      if (!parsedNPCIDs.isEmpty()) {
        parsedNPCIDs.clear();
      }
      String mobfilepath = root + ((Event)zonelist.get(i)).getMobPath() + ".dat";
      String eventfilepath = root + ((Event)zonelist.get(i)).getEventPath() + ".dat";
      

      parseMobListDAT(mobfilepath);
      

      parseEventListDAT(eventfilepath);
      

      Writer output = null;
      try
      {
        File file = new File(((Event)zonelist.get(i)).getName() + ".txt");
        output = new BufferedWriter(new FileWriter(file));
        output.write(fileout);
        output.close();
        System.out.println("Written file " + ((Event)zonelist.get(i)).getName() + ".txt");
      }
      catch (Exception e)
      {
        System.out.println("ERROR: Writing output file : " + e);
      }
      fileout = "";
    }
  }
  
  static void parseEventListDAT(String fpath)
  {
    try
    {
      File file = new File(fpath);
      InputStream is = new FileInputStream(file);
      
      long length = file.length();
      
      byte[] bytes = new byte[(int)length];
      
      int offset = 0;
      int numRead = 0;
      try
      {
        while ((offset < bytes.length) && ((numRead = is.read(bytes, offset, bytes.length - offset)) >= 0)) {
          offset += numRead;
        }
      }
      catch (Exception e)
      {
        throw e;
      }
      finally
      {
        is.close();
      }
      for (int k = 0; k < parsedNPCIDs.size(); k++) {
        for (int i = 0; i < bytes.length; i++) {
          if ((i + 3 < bytes.length) && 
            (bytes[i] == ((byte[])parsedNPCIDs.get(k))[0]) && (bytes[(i + 1)] == ((byte[])parsedNPCIDs.get(k))[1]) && (bytes[(i + 2)] == ((byte[])parsedNPCIDs.get(k))[2]) && (bytes[(i + 3)] == ((byte[])parsedNPCIDs.get(k))[3])) {
            if ((hexFor(bytes[(i + 5)]).equals("00")) && (hexFor(bytes[(i + 6)]).equals("00")) && (hexFor(bytes[(i + 7)]).equals("00")) && (hexFor(bytes[(i + 8)]).equals("00")) && (hexFor(bytes[(i + 9)]).equals("00")) && ((hexFor(bytes[(i + 10)]).equals("01")) || (hexFor(bytes[(i + 10)]).equals("04"))))
            {
              int numeventids = uInt(bytes[(i + 4)]);
              int startpos = 0;
              for (int l = i + 10; l < i + 65535; l++) {
                if ((hexFor(bytes[l]).equals("ff")) && (hexFor(bytes[(l + 1)]).equals("ff")))
                {
                  startpos = l;
                  break;
                }
              }
              if (startpos == 0)
              {
                System.out.println("Unable to find 0xFFFF for " + (String)parsedNPCNames.get(k));
                System.out.println("ID at 0d" + i + " with number of events: " + numeventids);
              }
              else
              {
                String hexid = hexFor(((byte[])parsedNPCIDs.get(k))[0]) + " " + hexFor(((byte[])parsedNPCIDs.get(k))[1]) + " " + hexFor(((byte[])parsedNPCIDs.get(k))[2]) + " " + hexFor(((byte[])parsedNPCIDs.get(k))[3]);
                


                fileout = fileout + "==========\r\n" + (String)parsedNPCNames.get(k) + " ID: " + hexid + "\r\n";
                fileout = fileout + "Number of Event IDs:" + numeventids + "\r\n";
                
                int reservedids = 0;
                for (int l = startpos; l < 2 * numeventids + startpos; l += 2) {
                  if ((hexFor(bytes[l]).equals("ff")) && (hexFor(bytes[(l + 1)]).equals("ff"))) {
                    reservedids++;
                  } else {
                    fileout = fileout + "0x" + hexFor(bytes[(l + 1)]) + hexFor(bytes[l]) + "  ";
                  }
                }
                fileout = fileout + " and " + reservedids + " reserved IDs.\r\n";
              }
            }
          }
        }
      }
    }
    catch (Exception ex)
    {
      System.out.println("ERROR: Event List DAT :" + ex);
    }
  }
  
  static void parseMobListDAT(String fpath)
  {
    try
    {
      File file = new File(fpath);
      InputStream is = new FileInputStream(file);
      
      long length = file.length();
      
      int numentries = (int)length / 32;
      System.out.println(fpath + " has " + numentries + " NPC entries.");
      
      byte[] bytes = new byte[(int)length];
      
      int offset = 0;
      int numRead = 0;
      try
      {
        while ((offset < bytes.length) && ((numRead = is.read(bytes, offset, bytes.length - offset)) >= 0)) {
          offset += numRead;
        }
      }
      catch (Exception e)
      {
        throw e;
      }
      finally
      {
        is.close();
      }
      for (int i = 0; i < bytes.length; i += 32)
      {
        String npcname = "";
        byte[] npcid = new byte[4];
        

        int terminator = -1;
        for (int j = 0; j < 28; j++) {
          if (bytes[(i + j)] == 0)
          {
            terminator = i + j;
            break;
          }
        }
        if (terminator == -1)
        {
          System.out.println("Can't find terminator!");
        }
        else
        {
          byte[] namebytes = new byte[terminator - i];
          for (int j = 0; j < terminator - i; j++) {
            namebytes[j] = bytes[(i + j)];
          }
          npcname = new String(namebytes);
        }
        npcid[0] = bytes[(i + 28)];
        npcid[1] = bytes[(i + 29)];
        npcid[2] = bytes[(i + 30)];
        npcid[3] = bytes[(i + 31)];
        
        boolean addEntry = true;
        if (npcname.equals("")) {
          addEntry = false;
        }
        if ((npcid[1] == 0) && (npcid[2] == 0) && (npcid[3] == 0)) {
          addEntry = false;
        }
        if (addEntry)
        {
          parsedNPCIDs.add(npcid);
          parsedNPCNames.add(npcname);
        }
      }
    }
    catch (Exception ex)
    {
      System.out.println("ERROR: Monster List DAT :" + ex);
    }
  }
  
  public static int uInt(byte b)
  {
    return b & 0xFF;
  }
  
  public static String hexFor(byte b)
  {
    int uint = uInt(b);
    String hex = Integer.toHexString(uint);
    if (hex.length() == 1) {
      hex = "0" + hex;
    }
    return hex;
  }
}
